import numpy as np
import cv2

# Read images
img1 = cv2.imread("../Wrinkles.jpg")
img2 = cv2.imread("../Face.jpg")

# Load landmarks
def loadLandmarks(filename):
    points = []
    with open(filename, "r") as f:
        for line in f.readlines():
            points.append((int(line.split()[1]), int(line.split()[2])))
    # print(len(points))
    return points
#    
points1 = loadLandmarks("../Wrinkles.txt")
points2 = loadLandmarks("../Face.txt")

# Display landmarks
u=0
def displayPoints(image,points):
    global u
    u+=1
    img = np.copy(image)
    for point in points:
        cv2.circle(img,point,int(2*img.shape[0]/300),(0,0,255),cv2.FILLED)
    cv2.imshow("landmarks"+str(u), cv2.resize(img,(int(img.shape[1]*300/img.shape[0]),300)))
    cv2.waitKey(0)
#
displayPoints(img1, points1)
displayPoints(img2, points2)

# Add forehead points
def forehead(points):
    brows = [25, 23, 20, 18] 
    eyes = [45, 47, 40, 36]
    return [ 
        (points[brows[i]][0]+2*(points[brows[i]][0]-points[eyes[i]][0]),
        points[brows[i]][1]+2*(points[brows[i]][1]-points[eyes[i]][1]))
        for i in range(4)
    ]
#
points1 += forehead(points1)
points2 += forehead(points2)
displayPoints(img1, points1)
displayPoints(img2, points2)

# Find triangulation
rect = (0,0,img2.shape[1],img2.shape[0])
delaunay = cv2.Subdiv2D(rect)
delaunay.insert(points2)
triangles = [ [(t[0],t[1]),(t[2],t[3]),(t[4],t[5])] for t in delaunay.getTriangleList() ]
#
localizer = dict()
for index, point in enumerate(points2):
    localizer[(float(point[0]),float(point[1]))] = index
#
triangulation = []
for triangle in triangles:
    indices = [ localizer[point] for point in triangle ]
    triangulation.append(indices)
#
def displayTriangulation(image,points,triangulation):
    global u
    u+=1
    img = np.copy(image)
    for indices in triangulation:
        p = [ points[index] for index in indices ]
        cv2.line(img,p[0],p[1],(0,0,255),int(2*img.shape[0]/300))
        cv2.line(img,p[1],p[2],(0,0,255),int(2*img.shape[0]/300))
        cv2.line(img,p[2],p[0],(0,0,255),int(2*img.shape[0]/300))
    cv2.imshow("triangulation"+str(u), cv2.resize(img,(int(img.shape[1]*300/img.shape[0]),300)))
    cv2.waitKey(0)
#
displayTriangulation(img2,points2,triangulation)
displayTriangulation(img1,points1,triangulation)

# Morphing: Warp img1 onto img2
def warpTriangle(image1,image2,triangle1,triangle2):
    rect1 = cv2.boundingRect(np.array(triangle1))
    rect2 = cv2.boundingRect(np.array(triangle2))
    offset1 = np.array([rect1[0],rect1[1]],np.float32)
    offset2 = np.array([rect2[0],rect2[1]],np.float32)
    subimage1 = image1[rect1[1]:rect1[1]+rect1[3],rect1[0]:rect1[0]+rect1[2]]
    subimage2 = image2[rect2[1]:rect2[1]+rect2[3],rect2[0]:rect2[0]+rect2[2]]
    transform = cv2.getAffineTransform(np.float32(triangle1)-offset1, np.float32(triangle2)-offset2)
    warped = cv2.warpAffine(subimage1,transform,(subimage2.shape[1],subimage2.shape[0]),None,flags=cv2.INTER_LINEAR,borderMode=cv2.BORDER_REFLECT_101)
    mask2 = np.zeros_like(subimage2)
    polygon = (np.float32(triangle2)-offset2).astype(np.int32)
    cv2.fillPoly(mask2, [polygon.astype(np.int32)], (255,255,255))
    subimage2[mask2>0] = warped[mask2>0]
#
img1Warped = np.copy(img2)
for indices in triangulation:
    triangle1 = [ points1[index] for index in indices ]
    triangle2 = [ points2[index] for index in indices ]
    warpTriangle(img1,img1Warped,triangle1,triangle2)
#
displayTriangulation(img1Warped,points2,triangulation)
cv2.imshow("warped",img1Warped)
cv2.waitKey(0)

# Mixed seemless cloning
mask = np.zeros_like(img2)
hull = cv2.convexHull(np.array(points2), returnPoints=True)
cv2.drawContours(mask,[hull],0,(255,255,255),cv2.FILLED)
cv2.imshow("mask",mask)
cv2.waitKey(0)
r = cv2.boundingRect(hull)
pivot = (r[0]+r[2]//2,r[1]+r[3]//2)
cloned = cv2.seamlessClone(img1Warped, img2, mask, pivot, cv2.MIXED_CLONE)
cv2.imshow("cloned",cloned)
cv2.waitKey(0)

cv2.destroyAllWindows()
